import numpy as np
from gammapy.datasets import MapDataset


class AsimovMapDataset(MapDataset):
    """
    AsimovMapDataset is a subclass of a gammapy MapDataset
    and provides asimov-like fake method.
    """

    def fake(self):
        """
        This method generates Asimov like counts,
        i.e. the counts are not drawn from a poisson distribution.
        """
        npred = self.npred()
        data = np.nan_to_num(npred.data, copy=True, nan=0.0, posinf=0.0, neginf=0.0)
        npred.data = data
        self.counts = npred



def CopyModelError(AttributeError):
    """Something went wrong during copying a model."""


def calc_ts_toyMC(dataset, test_statistic, poi_val, poi_true_val, poi_name):
    """Evaluates given test statistic for toy MCs.

    The given dataset will be copied and a new measurement will be simulated.
    The test statistic is evaluated for this dataset.
    """
    toy_dataset = copy_dataset_with_models(dataset)
    toy_dataset.models.parameters[poi_name].value = poi_true_val
    toy_dataset.fake()

    ts = test_statistic(toy_dataset, poi_name)
    return ts.evaluate(poi_val)


def copy_dataset_with_models(dataset):
    """Copies a dataset inlcuding the models."""
    dataset_copy = dataset.copy()
    copy_models_to_dataset(dataset.models, dataset_copy)

    return dataset_copy


def copy_models_to_dataset(models, dataset):
    """Copies models and assigns them to dataset."""
    model_copies = models.copy()
    for model in model_copies:
        if hasattr(model, "_name"):
            model._name = f"{dataset.name}-{model._name}"
        elif hasattr(model, "datasets_names"):
            model.datasets_names = [f"{dataset.name}"]
        else:
            raise CopyModelError(
                f"{model.__class__.__name__} doesn't provided `._name`"
                f"nor `.datasets_names`."
            )

    dataset.models = model_copies



import os

import astropy.constants as const
import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from gammapy.astro.darkmatter import (
    DarkMatterAnnihilationSpectralModel,
    JFactory,
    profiles,
)
from gammapy.data import Observation, observatory_locations
from gammapy.datasets import MapDataset
from gammapy.irf import load_cta_irfs
from gammapy.makers import MapDatasetMaker, SafeMaskMaker
from gammapy.maps import MapAxis, WcsGeom, WcsNDMap
from gammapy.modeling.models import (
    FoVBackgroundModel,
    Models,
    SkyModel,
    TemplateSpatialModel,
)


@pytest.fixture(scope="module")
def irfs():
    path = (
        f"{os.environ.get('GAMMAPY_DATA')}"
        f"/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )

    irfs = load_cta_irfs(path)

    return irfs


@pytest.fixture(scope="module")
def observation(irfs):
    livetime = 50 * u.hr
    pointing = SkyCoord(0, 0, unit="deg", frame="galactic")
    location = observatory_locations["cta_south"]
    obs = Observation.create(
        pointing=pointing, livetime=livetime, irfs=irfs, location=location
    )

    return obs


@pytest.fixture(scope="module")
def energy_axes():
    energy_reco = MapAxis.from_edges(
        np.logspace(-1.0, 1.0, 10), unit="TeV", name="energy", interp="log"
    )
    energy_axis_true = MapAxis.from_energy_bounds(
        "0.03 TeV", "300 TeV", nbin=20, per_decade=True, name="energy_true"
    )
    migra_axis = MapAxis.from_bounds(0.5, 2, nbin=150, node_type="edges", name="migra")

    return {"true": energy_axis_true, "reco": energy_reco, "migra": migra_axis}


@pytest.fixture(scope="module")
def geometry3d(observation, energy_axes):
    geom = WcsGeom.create(
        skydir=observation.pointing.fixed_icrs,
        width=(4, 4),
        binsz=0.02,
        frame="galactic",
        axes=[energy_axes["reco"]],
    )

    return geom


@pytest.fixture(scope="module")
def geometry2d(observation):
    geom = WcsGeom.create(
        skydir=observation.pointing.fixed_icrs,
        width=(4, 4),
        binsz=0.02,
        frame="galactic",
    )

    return geom


@pytest.fixture(scope="module")
def ursa_major_ii_profile():
    rhos = (
        10**-1.1331 * const.M_sun.to(u.GeV, equivalencies=u.mass_energy()) / u.pc**3
    )
    rs = 10**3.6317 * u.pc

    profile = profiles.NFWProfile(r_s=rs, rho_s=rhos)
    profile.DISTANCE_GC = 32 * u.kpc

    return profile


@pytest.fixture(scope="module")
def dm_models(observation, geometry2d, ursa_major_ii_profile):
    jfactory = JFactory(
        geom=geometry2d,
        profile=ursa_major_ii_profile,
        distance=ursa_major_ii_profile.DISTANCE_GC,
    )
    jfactor = jfactory.compute_differential_jfactor()
    jfact_map = WcsNDMap(geom=geometry2d, data=jfactor.value, unit=jfactor.unit)
    spatial_model = TemplateSpatialModel(jfact_map, normalize=False)

    spectral_model = DarkMatterAnnihilationSpectralModel(mass=50 * u.TeV, channel="b")

    model_simu = SkyModel(
        spatial_model=spatial_model,
        spectral_model=spectral_model,
        name="darkmatter",
    )

    bkg_model = FoVBackgroundModel(dataset_name="foo")

    models = Models([model_simu, bkg_model])

    return models


@pytest.fixture(scope="module")
def measurement_dataset(geometry3d, energy_axes, observation, dm_models):
    from titrate.utils import copy_models_to_dataset

    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    maker_safe_mask = SafeMaskMaker(methods=["offset-max"], offset_max=4.0 * u.deg)

    empty_measurement = MapDataset.create(
        geometry3d,
        energy_axis_true=energy_axes["true"],
        migra_axis=energy_axes["migra"],
        name="measurement",
    )

    measurement_dataset = maker.run(empty_measurement, observation)
    measurement_dataset = maker_safe_mask.run(measurement_dataset, observation)

    copy_models_to_dataset(dm_models, measurement_dataset)

    measurement_dataset.fake(random_state=42)

    return measurement_dataset


@pytest.fixture(scope="module")
def nosignal_dataset(geometry3d, energy_axes, observation, dm_models):
    from titrate.utils import copy_models_to_dataset

    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    maker_safe_mask = SafeMaskMaker(methods=["offset-max"], offset_max=4.0 * u.deg)

    empty_measurement = MapDataset.create(
        geometry3d,
        energy_axis_true=energy_axes["true"],
        migra_axis=energy_axes["migra"],
        name="measurement",
    )

    nosignal_dataset = maker.run(empty_measurement, observation)
    nosignal_dataset = maker_safe_mask.run(nosignal_dataset, observation)

    copy_models_to_dataset(dm_models, nosignal_dataset)
    nosignal_dataset.models.parameters["scale"].value = 0

    nosignal_dataset.fake(random_state=1337)

    return nosignal_dataset


@pytest.fixture(scope="module")
def asimov_dataset(geometry3d, energy_axes, observation, dm_models):
    from titrate.datasets import AsimovMapDataset
    from titrate.utils import copy_models_to_dataset

    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    maker_safe_mask = SafeMaskMaker(methods=["offset-max"], offset_max=4.0 * u.deg)

    empty_asimov = AsimovMapDataset.create(
        geometry3d,
        energy_axis_true=energy_axes["true"],
        migra_axis=energy_axes["migra"],
        name="asimov",
    )

    asimov_dataset = maker.run(empty_asimov, observation)
    asimov_dataset = maker_safe_mask.run(asimov_dataset, observation)

    copy_models_to_dataset(dm_models, asimov_dataset)

    asimov_dataset.fake()

    return asimov_dataset



import abc

import numpy as np
from gammapy.modeling import Fit

from titrate.datasets import AsimovMapDataset


class POIError(IndexError):
    """Parameter of interest is not defined in model"""


class AsimovApproximationError(IndexError):
    """Approimation on AsimovMapDataset fails"""


class TestStatistic(abc.ABC):
    r"""TestStatistic abstract base class.

    This class will build the base class for other test statistics like $q_\mu$
    or $\tilde{q}_\mu$.

    It contains two basic methods:
    1. `evaluate`, which takes a dataset with models to calculate the test statistic.
    2. `asymptotic_approximation`, which evaluates the distribution of the
        test statistic for a given value of the test statistic.
    """

    @abc.abstractmethod
    def evaluate(self):
        pass

    @abc.abstractmethod
    def asympotic_approximation(self):
        pass


class QMuTestStatistic(TestStatistic):
    def __init__(self, dataset, poi_name=""):
        self.dataset = dataset

        self.poi_name = poi_name
        try:
            self.dataset.models.parameters[self.poi_name]
        except IndexError:
            raise POIError(
                f"The parameter of interest `{self.poi_name}` is not available. "
                f"Please choose one from the following list: {self.check_for_pois()}"
            )

        self.fit = Fit()
        self.fit_result = self.fit.run(datasets=[self.dataset])
        self.poi_best = self.fit_result.parameters[self.poi_name].value
        self.likelihood_minimum = self.dataset.stat_sum()

    def evaluate(self, poi_val):
        """
        Computes the test statistic for a given dataset
        and parameter of interest (POI).
        """

        if self.poi_best > poi_val:
            return np.array([0])

        self.dataset.models.parameters[self.poi_name].scan_values = [poi_val]
        stats = self.fit.stat_profile(self.dataset, self.poi_name, reoptimize=True)
        ts = stats["stat_scan"] - self.likelihood_minimum

        return ts

    def check_for_pois(self):
        """POI must be a norm parameter by definition since
        we want to calculate TS for strength parameter of the signal.
        """
        pois = []
        for parameter in self.dataset.models.parameters:
            if parameter.is_norm:
                pois.append(parameter.name)
        return pois

    def sigma(self):
        if not isinstance(self.dataset, AsimovMapDataset):
            raise AsimovApproximationError(
                "`dataset` must be an `AsimovMapDataset` in order to calculate"
                " `sigma`"
            )
        return np.sqrt(self.fit_result.covariance_result.matrix[0, 0])

    def asympotic_approximation(self, ts_val, poi_val, poi_true_val):
        if not isinstance(self.dataset, AsimovMapDataset):
            raise AsimovApproximationError(
                "`dataset` must be an `AsimovMapDataset` in order to use the"
                " `asympotic_approximation`"
            )
        return (
            1
            / (2 * np.sqrt(2 * np.pi * ts_val))
            * np.exp(
                -0.5 * (np.sqrt(ts_val) - (poi_val - poi_true_val) / self.sigma()) ** 2
            )
        )


class QTildeMuTestStatistic(TestStatistic):
    def __init__(self, dataset, poi_name=""):
        self.dataset = dataset

        self.poi_name = poi_name
        try:
            self.dataset.models.parameters[self.poi_name]
        except IndexError:
            raise POIError(
                f"The parameter of interest `{self.poi_name}` is not available. "
                f"Please choose one from the following list: {self.check_for_pois()}"
            )

        self.fit = Fit()
        self.fit_result = self.fit.run(datasets=[self.dataset])
        self.poi_best = self.fit_result.parameters[self.poi_name].value
        if self.poi_best < 0:
            self.dataset.models.parameters[self.poi_name].scan_values = [0]
            self.likelihood_constant = self.fit.stat_profile(
                self.dataset, self.poi_name, reoptimize=True
            )["stat_scan"]
        else:
            self.likelihood_constant = self.dataset.stat_sum()

    def evaluate(self, poi_val):
        """
        Computes the test statistic for a given dataset
        and parameter of interest (POI).
        """

        if self.poi_best > poi_val:
            return np.array([0])

        self.dataset.models.parameters[self.poi_name].scan_values = [poi_val]
        stats = self.fit.stat_profile(self.dataset, self.poi_name, reoptimize=True)
        ts = stats["stat_scan"] - self.likelihood_constant

        return ts

    def check_for_pois(self):
        """POI must be a norm parameter by definition since
        we want to calculate TS for strength parameter of the signal.
        """
        pois = []
        for parameter in self.dataset.models.parameters:
            if parameter.is_norm:
                pois.append(parameter.name)
        return pois

    def sigma(self):
        if not isinstance(self.dataset, AsimovMapDataset):
            raise AsimovApproximationError(
                "`dataset` must be an `AsimovMapDataset` in order to calculate"
                " `sigma`"
            )
        return np.sqrt(self.fit_result.covariance_result.matrix[0, 0])

    def asympotic_approximation(self, ts_val, poi_val, poi_true_val):
        if not isinstance(self.dataset, AsimovMapDataset):
            raise AsimovApproximationError(
                "`dataset` must be an `AsimovMapDataset` in order to use the"
                " `asympotic_approximation`"
            )

        sigma = self.sigma()
        if ts_val > poi_val**2 / sigma**2:
            return (
                1
                / (np.sqrt(2 * np.pi) * 2 * poi_val / sigma)
                * np.exp(
                    -0.5
                    * (
                        ts_val
                        - (poi_val**2 - 2 * poi_val * poi_true_val) / sigma**2
                    )
                    ** 2
                    / (2 * poi_val / sigma) ** 2
                )
            )
        else:
            return (
                1
                / (2 * np.sqrt(2 * np.pi * ts_val))
                * np.exp(
                    -0.5 * (np.sqrt(ts_val) - (poi_val - poi_true_val) / sigma) ** 2
                )
            )
